"""This module enables cancellation of all currently running skills."""
from misty2py.utils.utils import get_misty
from typing import Dict
from misty2py.robot import Misty
from misty2py.utils.status import ActionLog
from misty2py.response import success_of_action_list


def cancel_skills(misty: Misty) -> Dict:
    """Cancels all skills currently running on Misty.

    Args:
        misty (Misty): The Misty on which to cancel running skills.

    Returns:
        Dict: a dictionary with the key `overall_success` specifying whether all actions were successful and a key for every action containing the dictionarised Misty2pyResponse.
    """
    actions = ActionLog()
    data = misty.get_info("skills_running").parse_to_dict()
    actions.append_({"get_running_skills": data})

    data = data.get("rest_response", {})
    result = data.get("result", [])
    to_cancel = []
    for dct in result:
        uid = dct.get("uniqueId", "")
        if len(uid) > 0:
            to_cancel.append(uid)

    for skill in to_cancel:
        data = misty.perform_action(
            "skill_cancel", data={"Skill": skill}
        ).parse_to_dict()
        actions.append_({"cancel_%s" % skill: data})

    return success_of_action_list(actions.get_())



"""Misty2py is a Python package for Misty II development using [Misty's REST API](https://docs.mistyrobotics.com/misty-ii/rest-api/api-reference/).

## Features
Misty2py can be used to:

- **perform actions** via sending a `POST` or `DELETE` requests to Misty's REST API;
- **obtain information** via sending a `GET` request to Misty's REST API;
- **receive streams of data** via subscribing to event types on Misty's Websockets API.

Misty2py uses following concepts:

- **action keywords** - keywords for endpoints of Misty's REST API that correspond to performing actions;
- **information keywords** - keywords for endpoints of Misty's REST API that correspond to retrieving information;
- **data shortcuts** - keywords for commonly used data that is supplied to Misty's API as the body of a `POST` request.

## Installation

### Poetry

To install misty2py, run `pip install misty2py`.

### From source

- If this is your first time using `misty2py` from source, do following:

  - Get Poetry (`python -m pip install poetry`) if you do not have it yet.
  - Copy `.env.example` to `.env`.
  - Replace the placeholder values in the new `.env` file.
  - Run `poetry install` to obtain all dependencies.

- Run the desired script via `poetry run python -m [name]` where `[name]` is the placeholder for the module location (in Python notation).
- If the scripts run but your Misty does not seem to respond, you have most likely provided an incorrect IP address for `MISTY_IP_ADDRESS` in `.env`.
- Pytests can be run via `poetry run pytest .`.
- The coverage report can be obtained via `poetry run pytest --cov-report html --cov=misty2py tests` for HTML output or via `poetry run pytest --cov=misty2py tests` for terminal output.

## Contribute

This project is not currently open for contributions. However, you can report issues via the [Issues Tracker](https://github.com/ChrisScarred/misty2py/issues) and inspect the [Source Code](https://github.com/ChrisScarred/misty2py).

## License

This project is licensed under the MIT License.

## Usage

### Getting started

The main object of this package is `Misty`, which is an abstract representation of Misty the robot. To initialise this object, it is required to know the IP address of the Misty robot that should be used.

The most direct way to initialise a `Misty` object is to use the IP address directly, which allows the user to get the object in one step via:

``` python
from misty2py.robot import Misty

my_misty = Misty("192.168.0.1")  #example IP address
```

This may be impractical and potentially even unsafe, so it is recommended to create a .env file in the project's directory, specify the IP address there via `MISTY_IP_ADDRESS="[ip_address_here]"` and use Misty2py's `EnvLoader` to load the IP address via:


``` python
from misty2py.robot import Misty
from misty2py.utils.env_loader import EnvLoader

env_loader = EnvLoader()
my_misty = Misty(env_loader.get_ip())
```

Assuming a `Misty` object called `my_misty` was obtained, all required actions can be performed via the following three methods:

``` python
# Performing an action (a POST or DELETE request):
my_misty.perform_action("<action_keyword>")

# Obtaining information (a GET request):
my_misty.get_info("<information_keyword>")

# Event related methods 
# (subscribing to an event, getting event data
# or event log and unsubscribing from an event):
my_misty.event("<parameter>")
```

### Responses

Any action performed via Misty2py which contains communication with Misty's APIs returns the `Misty2pyResponse` object. `Misty2pyResponse` is a uniform representation of two sub-responses that are present in any HTTP or WebSocket communication with Misty's APIs using Misty2py. The first sub-response is always from Misty2py and is represented by the attributes `Misty2pyResponse.misty2py_status` (`True` if no Misty2py-related errors were encountered) and potentially empty `Misty2pyResponse.error_msg` and `Misty2pyResponse.error_type` that contain error information if a Misty2py-related error was encountered. The other sub-response is either from Misty's REST API or Misty's WebSocket API. In the first case, it is represented by the attribute `Misty2pyResponse.rest_response` (Dict), and in the second case, it is represented by the attribute `Misty2pyResponse.ws_response`. One of these is always empty, because no action in Misty2py includes simultaneous communication with both APIs. For convenience, a `Misty2pyResponse` object can be easily parsed to a dictionary via the method `Misty2pyResponse.parse_to_dict`.

### Obtaining information

Obtaining digital information is handled by `misty2py.robot::get_info` method which has two arguments. The argument `info_name` is required and it specifies the string information keyword corresponding to an endpoint in Misty's REST API. The argument `params` is optional and it supplies a dictionary of parameter name and parameter value pairs. This argument defaults to `{}` (an empty dictionary).

### Performing actions

Performing physical and digital actions including removal of non-system files is handled by `misty2py.robot::perform_action()` method which takes two arguments. The argument `action_name` is required and it specifies the string action keyword corresponding to an endpoint in Misty’s REST API. The second argument, `data`, is optional and it specifies the data to pass to the request as a dictionary or a data shortcut (string). The `data` argument defaults to `{}` (an empty dictionary).

### Event types

Misty's WebSocket API follows PUB-SUB architecture, which means that in order to obtain event data in Misty's framework, it is required to **subscribe** to an event type on Misty's WebSocket API. The WebSocket server then streams data to the WebSocket client, which receives it a separate thread. To **access the data,** `misty2py.robot::event` method must be called with `"get_data"` parameter from the main thread. When the data are no longer required to be streamed to the client, an event type can be **unsubscribed** which both kills the event thread and stops the API from sending more data.

**Subscribing** to an event is done via `misty2py.robot::event` with the parameter `"subscribe"` and following keyword arguments:

- `type` - *required;* event type string as documented in [Event Types Docs](https://docs.mistyrobotics.com/misty-ii/robot/sensor-data/ "Misty Robotics Event Types").
- `name` - *optional;* a custom event name string; must be unique.
- `return_property` - *optional;* the property to return from Misty's websockets; all properties are returned if return_property is not supplied.
- `debounce` - *optional;* the interval in ms at which new information is sent; defaults to `250`.
- `len_data_entries` - *optional;* the maximum number of data entries to keep (discards in fifo style); defaults to `10`.
- `event_emitter` - *optional;* an event emitter function which emits an event upon message recieval. Supplies the message content as an argument.

**Accessing the data** of an event or its log is done via `misty2py.robot::event` with the parameter `"get_data"` or `"get_log"` and a keyword argument `name` (the name of the event).

**Unsubscribing** from an event is done via `misty2py.robot::event` with the parameter `"unsubscribe"` and a keyword argument `name` (the name of the event).

A bare-bones implementation of event subscription can be seen below.

```python
import time

from misty2py.robot import Misty
from misty2py.utils.env_loader import EnvLoader

env_loader = EnvLoader

m = Misty(env_loader.get_ip())

d = m.event("subscribe", type = "BatteryCharge")
e_name = d.get("event_name")

time.sleep(1)

d = m.event("get_data", name = e_name)

d = m.event("unsubscribe", name = e_name)
```

The following example shows a more realistic scenario which includes an event emitter and an event listener.

```python
import time
from pymitter import EventEmitter

from misty2py.robot import Misty
from misty2py.utils.env_loader import EnvLoader

env_loader = EnvLoader

m = Misty(env_loader.get_ip())
ee = EventEmitter()
event_name = "myevent_001"

@ee.on(event_name)
def listener(data):
    print(data)

d = m.event("subscribe", type = "BatteryCharge", 
            name = event_name, event_emitter = ee)

time.sleep(2)

d = m.event("unsubscribe", name = event_name)
```

### Adding custom keywords and shortcuts

Custom keywords and shortcuts can be passed to a Misty object while declaring a new instance by using the optional arguments `custom_info`, `custom_actions` and `custom_data`. 

The argument `custom_info` can be used to pass custom information keywords as a dictionary with keys being the information keywords and values being the endpoints. An information keyword can only be used for a `GET` method supporting endpoint.

The argument `custom_actions` can be used to pass custom action keywords as a dictionary with keys being the action keywords and values being a dictionary of an `"endpoint"` key (str) and a `"method"` key (str). The `"method"` values must be one of `post`, `delete`, `put`, `head`, `options` and `patch`. However, it should be noted that Misty's REST API currently only has `GET`, `POST` and `DELETE` methods. The rest of the methods was implement in Misty2py for forwards-compatibility.

The argument `custom_data` can be used to pass custom data shortcuts as a dictionary with keys being the data shortcuts and values being the dictionary of data values.

For futher illustration, an example of passing custom keywords and shortcuts can be seen below.

```python
custom_allowed_infos = {
    "hazards_settings": "api/hazards/settings"
}

custom_allowed_data = {
    "amazement": {
        "FileName": "s_Amazement.wav"
    },
    "red": {
        "red": "255",
        "green": "0",
        "blue": "0"
    }
}

custom_allowed_actions = {
    "audio_play" : {
        "endpoint" : "api/audio/play",
        "method" : "post"
    },
    "delete_audio" : {
        "endpoint" : "api/audio",
        "method" : "delete"
    }
}

misty_robot = Misty("0.0.0.0", 
    custom_info=custom_allowed_infos, 
    custom_actions=custom_allowed_actions, 
    custom_data=custom_allowed_data)
```


## Default Keywords and Shortcuts

<details> 
  <summary>List of supported action keywords</summary>

<ul>
<li><code>led</code> for <strong>post</strong> request to <code>api/led</code> endpoint</li>
<li><code>led_trans</code> for <strong>post</strong> request to <code>api/led/transition</code> endpoint</li>
<li><code>notification_settings</code> for <strong>post</strong> request to <code>api/notification/settings</code> endpoint</li>
<li><code>audio_upload</code> for <strong>post</strong> request <code>api/audio</code> to endpoint</li>
<li><code>audio_play</code> for <strong>post</strong> request to <code>api/audio/play</code> endpoint</li>
<li><code>audio_pause</code> for <strong>post</strong> request to <code>api/audio/pause</code> endpoint</li>
<li><code>audio_stop</code> for <strong>post</strong> request to <code>api/audio/stop</code> endpoint</li>
<li><code>audio_delete</code> for <strong>delete</strong> request to <code>api/audio</code> endpoint</li>
<li><code>audio_record_start</code> for <strong>post</strong> request to <code>api/audio/record/start</code> endpoint</li>
<li><code>audio_record_stop</code> for <strong>post</strong> request to <code>api/audio/record/stop</code> endpoint</li>
<li><code>audio_disable</code> for <strong>post</strong> request to <code>api/services/audio/disable</code> endpoint</li>
<li><code>audio_enable</code> for <strong>post</strong> request to <code>api/services/audio/enable</code> endpoint</li>
<li><code>image_upload</code> for <strong>post</strong> request to <code>api/images</code> endpoint</li>
<li><code>image_show</code> for <strong>post</strong> request to <code>api/images/display</code> endpoint</li>
<li><code>image_settings</code> for <strong>post</strong> request to <code>api/images/settings</code> endpoint</li>
<li><code>image_delete</code> for <strong>delete</strong> request to <code>api/images</code> endpoint</li>
<li><code>text_show</code> for <strong>post</strong> request to <code>api/text/display</code> endpoint</li>
<li><code>text_settings</code> for <strong>post</strong> request to <code>api/text/settings</code> endpoint</li>
<li><code>video_upload</code> for <strong>post</strong> request to <code>api/videos</code> endpoint</li>
<li><code>video_show</code> for <strong>post</strong> request to <code>api/videos/display</code> endpoint</li>
<li><code>video_settings</code> for <strong>post</strong> request to <code>api/videos/settings</code> endpoint</li>
<li><code>video_delete</code> for <strong>delete</strong> request to <code>api/videos</code> endpoint</li>
<li><code>blink_mapping_delete</code> for <strong>delete</strong> request to <code>api/blink/images</code> endpoint</li>
<li><code>slam_enable</code> for <strong>post</strong> request to <code>api/services/slam/enable</code> endpoint</li>
<li><code>slam_disable</code> for <strong>post</strong> request to <code>api/services/slam/disable</code> endpoint</li>
<li><code>slam_sensors_reset</code> for <strong>post</strong> request to <code>api/slam/reset</code> endpoint</li>
<li><code>slam_mapping_start</code> for <strong>post</strong> request to <code>api/slam/map/start</code> endpoint</li>
<li><code>slam_mapping_stop</code> for <strong>post</strong> request to <code>api/slam/map/stop</code> endpoint</li>
<li><code>slam_map_current</code> for <strong>post</strong> request to <code>api/slam/map/current</code> endpoint</li>
<li><code>slam_map_rename</code> for <strong>post</strong> request to <code>api/slam/map/rename</code> endpoint</li>
<li><code>slam_infrared_settings</code> for <strong>post</strong> request to <code>api/slam/settings/ir</code> endpoint</li>
<li><code>slam_visible_settings</code> for <strong>post</strong> request to <code>api/slam/settings/visible</code> endpoint</li>
<li><code>slam_map_delete</code> for <strong>delete</strong> request to <code>api/slam/map</code> endpoint</li>
<li><code>slam_docking_locate_start</code> for <strong>post</strong> request to <code>api/slam/docking/start</code> endpoint</li>
<li><code>slam_docking_locate_stop</code> for <strong>post</strong> request to <code>api/slam/docking/stop</code> endpoint</li>
<li><code>streaming_slam_start</code> for <strong>post</strong> request to <code>api/slam/streaming/start</code> endpoint</li>
<li><code>streaming_slam_stop</code> for <strong>post</strong> request to <code>api/slam/streaming/stop</code> endpoint</li>
<li><code>slam_track_start</code> for <strong>post</strong> request to <code>api/slam/track/start</code> endpoint</li>
<li><code>slam_track_stop</code> for <strong>post</strong> request to <code>api/slam/track/stop</code> endpoint</li>
<li><code>recording_start</code> for <strong>post</strong> request to <code>api/videos/recordings/start</code> endpoint</li>
<li><code>recording_stop</code> for <strong>post</strong> request to <code>api/videos/recordings/stop</code> endpoint</li>
<li><code>recording_rename</code> for <strong>post</strong> request to <code>api/videos/recordings/rename</code> endpoint</li>
<li><code>recording_delete</code> for <strong>delete</strong> request to <code>api/videos/recordings</code> endpoint</li>
<li><code>face_detection_start</code> for <strong>post</strong> request to <code>api/faces/detection/start</code> endpoint</li>
<li><code>face_detection_stop</code> for <strong>post</strong> request to <code>api/faces/detection/stop</code> endpoint</li>
<li><code>face_recognition_start</code> for <strong>post</strong> request to <code>api/faces/recognition/start</code> endpoint</li>
<li><code>face_recognition_stop</code> for <strong>post</strong> request to <code>api/faces/recognition/stop</code> endpoint</li>
<li><code>face_train_start</code> for <strong>post</strong> request to <code>api/faces/training/start</code> endpoint</li>
<li><code>face_train_cancel</code> for <strong>post</strong> request to <code>api/faces/training/cancel</code> endpoint</li>
<li><code>face_delete</code> for <strong>delete</strong> request to <code>api/faces</code> endpoint</li>
<li><code>skill_upload</code> for <strong>post</strong> request to <code>api/skills</code> endpoint</li>
<li><code>skill_start</code> for <strong>post</strong> request to <code>api/skills/start</code> endpoint</li>
<li><code>skills_reload</code> for <strong>post</strong> request to <code>api/skills/reload</code> endpoint</li>
<li><code>skill_load</code> for <strong>post</strong> request to <code>api/skills/load</code> endpoint</li>
<li><code>skill_cancel</code> for <strong>post</strong> request to <code>api/skills/cancel</code> endpoint</li>
<li><code>skill_delete</code> for <strong>delete</strong> request to <code>api/skills</code> endpoint</li>
<li><code>wifi_add</code> for <strong>post</strong> request to <code>api/networks/create</code> endpoint</li>
<li><code>wifi_connect</code> for <strong>post</strong> request to <code>api/networks</code> endpoint</li>
<li><code>wifi_delete</code> for <strong>delete</strong> request to <code>api/networks</code> endpoint</li>
<li><code>wifi_hotspot_start</code> for <strong>post</strong> request to <code>api/networks/hotspot/start</code> endpoint</li>
<li><code>wifi_hotspot_stop</code> for <strong>post</strong> request to <code>api/networks/hotspot/stop</code> endpoint</li>
<li><code>write_serial</code> for <strong>post</strong> request to <code>api/serial</code> endpoint</li>
<li><code>event_listener</code> for <strong>post</strong> request to <code>api/skills/event</code> endpoint</li>
<li><code>website_show</code> for <strong>post</strong> request to <code>api/webviews/display</code> endpoint</li>
<li><code>website_settings</code> for <strong>post</strong> request to <code>api/webviews/settings</code> endpoint</li>
<li><code>blink_on</code> for <strong>post</strong> request to <code>api/blink</code> endpoint</li>
<li><code>blink_settings</code> for <strong>post</strong> request to <code>api/blink/settings</code> endpoint</li>
<li><code>display_settings</code> for <strong>post</strong> request to <code>api/display/settings</code> endpoint</li>
<li><code>flashlight_on</code> for <strong>post</strong> request to <code>api/flashlight</code> endpoint</li>
<li><code>speak</code> for <strong>post</strong> request to <code>api/tts/speak</code> endpoint</li>
<li><code>speak_stop</code> for <strong>post</strong> request to <code>api/tts/stop</code> endpoint</li>
<li><code>speech_capture</code> for <strong>post</strong> request to <code>api/audio/speech/capture</code> endpoint</li>
<li><code>drive</code> for <strong>post</strong> request to <code>api/drive</code> endpoint</li>
<li><code>drive_arc</code> for <strong>post</strong> request to <code>api/drive/arc</code> endpoint</li>
<li><code>drive_heading</code> for <strong>post</strong> request to <code>api/drive/hdt</code> endpoint</li>
<li><code>drive_time</code> for <strong>post</strong> request to <code>api/drive/time</code> endpoint</li>
<li><code>drive_track</code> for <strong>post</strong> request to <code>api/drive/track</code> endpoint</li>
<li><code>drive_stop</code> for <strong>post</strong> request to <code>api/drive/stop</code> endpoint</li>
<li><code>drive_to_loc</code> for <strong>post</strong> request to <code>api/drive/coordinates</code> endpoint</li>
<li><code>drive_on_path</code> for <strong>post</strong> request to <code>api/drive/path</code> endpoint</li>
<li><code>halt</code> for <strong>post</strong> request to <code>api/halt</code> endpoint</li>
<li><code>arm_move</code> for <strong>post</strong> request to <code>api/arms</code> endpoint</li>
<li><code>arms_move</code> for <strong>post</strong> request to <code>api/arms/set</code> endpoint</li>
<li><code>head_move</code> for <strong>post</strong> request to <code>api/head</code> endpoint</li>
<li><code>hazard_settings</code> for <strong>post</strong> request to <code>api/hazard/updatebasesettings</code> endpoint</li>
<li><code>streaming_av_start</code> for <strong>post</strong> request to <code>api/avstreaming/start</code> endpoint</li>
<li><code>streaming_av_stop</code> for <strong>post</strong> request to <code>api/avstreaming/stop</code> endpoint</li>
<li><code>streaming_av_disable</code> for <strong>post</strong> request to <code>api/services/avstreaming/disable</code> endpoint</li>
<li><code>streaming_av_enable</code> for <strong>post</strong> request to <code>api/services/avstreaming/enable</code> endpoint</li>
<li><code>keyphrase_recognition_start</code> for <strong>post</strong> request to <code>api/audio/keyphrase/start</code> endpoint</li>
<li><code>keyphrase_recognition_stop</code> for <strong>post</strong> request to <code>api/audio/keyphrase/stop</code> endpoint</li>
<li><code>update_allow</code> for <strong>post</strong> request to <code>api/system/update/allow</code> endpoint</li>
<li><code>update_perform</code> for <strong>post</strong> request to <code>api/system/update</code> endpoint</li>
<li><code>update_perform_targeted</code> for <strong>post</strong> request to <code>api/system/update/component</code> endpoint</li>
<li><code>update_prevent</code> for <strong>post</strong> request to <code>api/system/update/prevent</code> endpoint</li>
<li><code>error_text_clear</code> for <strong>post</strong> request to <code>api/text/error/clear</code> endpoint</li>
<li><code>camera_disable</code> for <strong>post</strong> request to <code>api/services/camera/disable</code> endpoint</li>
<li><code>camera_enable</code> for <strong>post</strong> request to <code>api/services/camera/enable</code> endpoint</li>
<li><code>restart</code> for <strong>post</strong> request to <code>api/reboot</code> endpoint</li>
<li><code>volume_settings</code> for <strong>post</strong> request to <code>api/audio/volume</code> endpoint</li>
<li><code>logs_settings</code> for <strong>post</strong> request to <code>api/logs/level</code> endpoint</li>
<li><code>websocket_settings</code> for <strong>post</strong> request to <code>api/websocket/version</code> endpoint</li>
<li><code>external_request</code> for <strong>post</strong> request to <code>api/request</code> endpoint</li>
</ul>
</details>

<details> 
  <summary>List of supported information keywords</summary>

<ul>
<li><code>audio_file</code> for <strong>get</strong> request to <code>api/audio</code> endpoint</li>
<li><code>audio_list</code> for <strong>get</strong> request to <code>api/audio/list</code> endpoint</li>
<li><code>audio_status</code> for <strong>get</strong> request to <code>api/services/audio</code> endpoint</li>
<li><code>image_file</code> for <strong>get</strong> request to <code>api/images</code> endpoint</li>
<li><code>image_list</code> for <strong>get</strong> request to <code>api/images/list</code> endpoint</li>
<li><code>video_file</code> for <strong>get</strong> request to <code>api/videos</code> endpoint</li>
<li><code>video_list</code> for <strong>get</strong> request to <code>api/videos/list</code> endpoint</li>
<li><code>av_status</code> for <strong>get</strong> request to <code>api/services/avstreaming</code> endpoint</li>
<li><code>sensor_values</code> for <strong>get</strong> request to <code>api/serial</code> endpoint</li>
<li><code>map_file</code> for <strong>get</strong> request to <code>api/slam/map</code> endpoint</li>
<li><code>current_map_id</code> for <strong>get</strong> request to <code>api/slam/map/current</code> endpoint</li>
<li><code>map_id_list</code> for <strong>get</strong> request to <code>api/slam/map/ids</code> endpoint</li>
<li><code>slam_diagnostics</code> for <strong>get</strong> request to <code>api/slam/diagnostics</code> endpoint</li>
<li><code>slam_path</code> for <strong>get</strong> request to <code>api/slam/path</code> endpoint</li>
<li><code>slam_status</code> for <strong>get</strong> request to <code>api/slam/status</code> endpoint</li>
<li><code>slam_enabled</code> for <strong>get</strong> request to <code>api/services/slam</code> endpoint</li>
<li><code>picture_depth</code> for <strong>get</strong> request to <code>api/cameras/depth</code> endpoint</li>
<li><code>picture_fisheye</code> for <strong>get</strong> request to <code>api/cameras/fisheye</code> endpoint</li>
<li><code>picture_rgb</code> for <strong>get</strong> request to <code>api/cameras/rgb</code> endpoint</li>
<li><code>faces_known</code> for <strong>get</strong> request to <code>api/faces</code> endpoint</li>
<li><code>recording_file</code> for <strong>get</strong> request to <code>api/videos/recordings</code> endpoint</li>
<li><code>recording_list</code> for <strong>get</strong> request to <code>api/videos/recordings/list</code> endpoint</li>
<li><code>skills_running</code> for <strong>get</strong> request to <code>api/skills/running</code> endpoint</li>
<li><code>skills_known</code> for <strong>get</strong> request to <code>api/skills</code> endpoint</li>
<li><code>wifis_available</code> for <strong>get</strong> request to <code>api/networks/scan</code> endpoint</li>
<li><code>wifis_saved</code> for <strong>get</strong> request to <code>api/networks</code> endpoint</li>
<li><code>battery_status</code> for <strong>get</strong> request to <code>api/battery</code> endpoint</li>
<li><code>camera_status</code> for <strong>get</strong> request to <code>api/services/camera</code> endpoint</li>
<li><code>blink_settings</code> for <strong>get</strong> request to <code>api/blink/settings</code> endpoint</li>
<li><code>hazards_settings</code> for <strong>get</strong> request to <code>api/hazards/settings</code> endpoint</li>
<li><code>camera_settings</code> for <strong>get</strong> request to <code>api/camera</code> endpoint</li>
<li><code>slam_visible_settings</code> for <strong>get</strong> request to <code>api/slam/settings/visible</code> endpoint</li>
<li><code>slam_infrared_settings</code> for <strong>get</strong> request to <code>api/slam/settings/ir</code> endpoint</li>
<li><code>update_settings</code> for <strong>get</strong> request to <code>api/system/update/settings</code> endpoint</li>
<li><code>device</code> for <strong>get</strong> request to <code>api/device</code> endpoint</li>
<li><code>help</code> for <strong>get</strong> request to <code>api/help</code> endpoint</li>
<li><code>log</code> for <strong>get</strong> request to <code>api/logs</code> endpoint</li>
<li><code>log_level</code> for <strong>get</strong> request to <code>api/logs/level</code> endpoint</li>
<li><code>update_available</code> for <strong>get</strong> request to <code>api/system/updates</code> endpoint</li>
<li><code>websockets</code> for <strong>get</strong> request to <code>api/websockets</code> endpoint</li>
<li><code>websocket_version</code> for <strong>get</strong> request to <code>api/websocket/version</code></li>
</ul>
</details>

<details> 
  <summary>List of supported data shortcuts</summary>

<ul>
<li><code>led_off</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>white_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>red_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>green_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>blue_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>yellow_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>cyan_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>magenta_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>orange_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;125&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>lime_light</code> for <code>{&quot;red&quot;: &quot;125&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>aqua_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;125&quot;}</code></li>
<li><code>azure_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;125&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>violet_light</code> for <code>{&quot;red&quot;: &quot;125&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>pink_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;125&quot;}</code></li>
<li><code>low_volume</code> for <code>{&quot;Volume&quot;: &quot;5&quot;}</code></li>
<li><code>image_admiration</code> for <code>{&quot;FileName&quot;: &quot;e_Admiration.jpg&quot;}</code></li>
<li><code>image_aggressiveness</code> for <code>{&quot;FileName&quot;: &quot;e_Aggressiveness.jpg&quot;}</code></li>
<li><code>image_amazement</code> for <code>{&quot;FileName&quot;: &quot;e_Amazement.jpg&quot;}</code></li>
<li><code>image_anger</code> for <code>{&quot;FileName&quot;: &quot;e_Anger.jpg&quot;}</code></li>
<li><code>image_concerned</code> for <code>{&quot;FileName&quot;: &quot;e_ApprehensionConcerned.jpg&quot;}</code></li>
<li><code>image_contempt</code> for <code>{&quot;FileName&quot;: &quot;e_Contempt.jpg&quot;}</code></li>
<li><code>image_content_left</code> for <code>{&quot;FileName&quot;: &quot;e_ContentLeft.jpg&quot;}</code></li>
<li><code>image_content_right</code> for <code>{&quot;FileName&quot;: &quot;e_ContentRight.jpg&quot;}</code></li>
<li><code>image_content_default</code> for <code>{&quot;FileName&quot;: &quot;e_DefaultContent.jpg&quot;}</code></li>
<li><code>image_disgust</code> for <code>{&quot;FileName&quot;: &quot;e_Disgust.jpg&quot;}</code></li>
<li><code>image_disoriented</code> for <code>{&quot;FileName&quot;: &quot;e_Disoriented.jpg&quot;}</code></li>
<li><code>image_hilarious</code> for <code>{&quot;FileName&quot;: &quot;e_EcstacyHilarious.jpg&quot;}</code></li>
<li><code>image_starry_eyed</code> for <code>{&quot;FileName&quot;: &quot;e_EcstacyStarryEyed.jpg&quot;}</code></li>
<li><code>image_fear</code> for <code>{&quot;FileName&quot;: &quot;e_Fear.jpg&quot;}</code></li>
<li><code>image_grief</code> for <code>{&quot;FileName&quot;: &quot;e_Grief.jpg&quot;}</code></li>
<li><code>image_joy_1</code> for <code>{&quot;FileName&quot;: &quot;e_Joy.jpg&quot;}</code></li>
<li><code>image_joy_2</code> for <code>{&quot;FileName&quot;: &quot;e_Joy2.jpg&quot;}</code></li>
<li><code>image_goofy_1</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy.jpg&quot;}</code></li>
<li><code>image_goofy_2</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy2.jpg&quot;}</code></li>
<li><code>image_goofy_3</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy3.jpg&quot;}</code></li>
<li><code>image_love</code> for <code>{&quot;FileName&quot;: &quot;e_Love.jpg&quot;}</code></li>
<li><code>image_rage_1</code> for <code>{&quot;FileName&quot;: &quot;e_Rage.jpg&quot;}</code></li>
<li><code>image_rage_2</code> for <code>{&quot;FileName&quot;: &quot;e_Rage2.jpg&quot;}</code></li>
<li><code>image_rage_3</code> for <code>{&quot;FileName&quot;: &quot;e_Rage3.jpg&quot;}</code></li>
<li><code>image_rage_4</code> for <code>{&quot;FileName&quot;: &quot;e_Rage4.jpg&quot;}</code></li>
<li><code>image_remorse</code> for <code>{&quot;FileName&quot;: &quot;e_RemorseShame.jpg&quot;}</code></li>
<li><code>image_sadness</code> for <code>{&quot;FileName&quot;: &quot;e_Sadness.jpg&quot;}</code></li>
<li><code>image_sleping_1</code> for <code>{&quot;FileName&quot;: &quot;e_Sleeping.jpg&quot;}</code></li>
<li><code>image_sleeping_2</code> for <code>{&quot;FileName&quot;: &quot;e_SleepingZZZ.jpg&quot;}</code></li>
<li><code>image_sleepy_1</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy.jpg&quot;}</code></li>
<li><code>image_sleepy_2</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy2.jpg&quot;}</code></li>
<li><code>image_sleepy_3</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy3.jpg&quot;}</code></li>
<li><code>image_sleepy_4</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy4.jpg&quot;}</code></li>
<li><code>image_surprise</code> for <code>{&quot;FileName&quot;: &quot;e_Surprise.jpg&quot;}</code></li>
<li><code>image_system_black_screen</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlackScreen.jpg&quot;}</code></li>
<li><code>image_system_blink_large</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlinkLarge.jpg&quot;}</code></li>
<li><code>image_system_blink_standard</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlinkStandard.jpg&quot;}</code></li>
<li><code>image_system_camera</code> for <code>{&quot;FileName&quot;: &quot;e_SystemCamera.jpg&quot;}</code></li>
<li><code>image_system_flash</code> for <code>{&quot;FileName&quot;: &quot;e_SystemFlash.jpg&quot;}</code></li>
<li><code>image_system_gear_prompt</code> for <code>{&quot;FileName&quot;: &quot;e_SystemGearPrompt.jpg&quot;}</code></li>
<li><code>image_system_logo_prompt</code> for <code>{&quot;FileName&quot;: &quot;e_SystemLogoPrompt.jpg&quot;}</code></li>
<li><code>image_terror_1</code> for <code>{&quot;FileName&quot;: &quot;e_Terror.jpg&quot;}</code></li>
<li><code>image_terror_2</code> for <code>{&quot;FileName&quot;: &quot;e_Terror2.jpg&quot;}</code></li>
<li><code>image_terror_left</code> for <code>{&quot;FileName&quot;: &quot;e_TerrorLeft.jpg&quot;}</code></li>
<li><code>image_terror_right</code> for <code>{&quot;FileName&quot;: &quot;e_TerrorRight.jpg&quot;}</code></li>
<li><code>sound_acceptance</code> for <code>{&quot;FileName&quot;: &quot;s_Acceptance.wav&quot;}</code></li>
<li><code>sound_amazement_1</code> for <code>{&quot;FileName&quot;: &quot;s_Amazement.wav&quot;}</code></li>
<li><code>sound_amazement_2</code> for <code>{&quot;FileName&quot;: &quot;s_Amazement2.wav&quot;}</code></li>
<li><code>sound_anger_1</code> for <code>{&quot;FileName&quot;: &quot;s_Anger.wav&quot;}</code></li>
<li><code>sound_anger_2</code> for <code>{&quot;FileName&quot;: &quot;s_Anger2.wav&quot;}</code></li>
<li><code>sound_anger_3</code> for <code>{&quot;FileName&quot;: &quot;s_Anger3.wav&quot;}</code></li>
<li><code>sound_anger_4</code> for <code>{&quot;FileName&quot;: &quot;s_Anger4.wav&quot;}</code></li>
<li><code>sound_annoyance_1</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance.wav&quot;}</code></li>
<li><code>sound_annoyance_2</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance2.wav&quot;}</code></li>
<li><code>sound_annoyance_3</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance3.wav&quot;}</code></li>
<li><code>sound_annoyance_4</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance4.wav&quot;}</code></li>
<li><code>sound_awe_1</code> for <code>{&quot;FileName&quot;: &quot;s_Awe.wav&quot;}</code></li>
<li><code>sound_awe_2</code> for <code>{&quot;FileName&quot;: &quot;s_Awe2.wav&quot;}</code></li>
<li><code>sound_awe_3</code> for <code>{&quot;FileName&quot;: &quot;s_Awe3.wav&quot;}</code></li>
<li><code>sound_boredom</code> for <code>{&quot;FileName&quot;: &quot;s_Boredom.wav&quot;}</code></li>
<li><code>sound_disapproval</code> for <code>{&quot;FileName&quot;: &quot;s_Disapproval.wav&quot;}</code></li>
<li><code>sound_disgust_1</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust.wav&quot;}</code></li>
<li><code>sound_disgust_2</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust2.wav&quot;}</code></li>
<li><code>sound_disgust_3</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust3.wav&quot;}</code></li>
<li><code>sound_disoriented_1</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused.wav&quot;}</code></li>
<li><code>sound_disoriented_2</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused2.wav&quot;}</code></li>
<li><code>sound_disoriented_3</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused3.wav&quot;}</code></li>
<li><code>sound_disoriented_4</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused4.wav&quot;}</code></li>
<li><code>sound_disoriented_5</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused5.wav&quot;}</code></li>
<li><code>sound_disoriented_6</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused6.wav&quot;}</code></li>
<li><code>sound_distraction</code> for <code>{&quot;FileName&quot;: &quot;s_Distraction.wav&quot;}</code></li>
<li><code>sound_ecstacy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Ecstacy.wav&quot;}</code></li>
<li><code>sound_ecstacy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Ecstacy2.wav&quot;}</code></li>
<li><code>sound_fear</code> for <code>{&quot;FileName&quot;: &quot;s_Fear.wav&quot;}</code></li>
<li><code>sound_grief_1</code> for <code>{&quot;FileName&quot;: &quot;s_Grief.wav&quot;}</code></li>
<li><code>sound_grief_2</code> for <code>{&quot;FileName&quot;: &quot;s_Grief2.wav&quot;}</code></li>
<li><code>sound_grief_3</code> for <code>{&quot;FileName&quot;: &quot;s_Grief3.wav&quot;}</code></li>
<li><code>sound_grief_4</code> for <code>{&quot;FileName&quot;: &quot;s_Grief4.wav&quot;}</code></li>
<li><code>sound_joy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Joy.wav&quot;}</code></li>
<li><code>sound_joy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Joy2.wav&quot;}</code></li>
<li><code>sound_joy_3</code> for <code>{&quot;FileName&quot;: &quot;s_Joy3.wav&quot;}</code></li>
<li><code>sound_joy_4</code> for <code>{&quot;FileName&quot;: &quot;s_Joy4.wav&quot;}</code></li>
<li><code>sound_loathing</code> for <code>{&quot;FileName&quot;: &quot;s_Loathing.wav&quot;}</code></li>
<li><code>sound_love</code> for <code>{&quot;FileName&quot;: &quot;s_Love.wav&quot;}</code></li>
<li><code>sound_phrase_bye_bye</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseByeBye.wav&quot;}</code></li>
<li><code>sound_phrase_evil</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseEvilAhHa.wav&quot;}</code></li>
<li><code>sound_phrase_hello</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseHello.wav&quot;}</code></li>
<li><code>sound_phrase_no</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseNoNoNo.wav&quot;}</code></li>
<li><code>sound_phrase_oopsy</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOopsy.wav&quot;}</code></li>
<li><code>sound_phrase_ow</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOwOwOw.wav&quot;}</code></li>
<li><code>sound_phrase_oww</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOwwww.wav&quot;}</code></li>
<li><code>sound_phrase_uh</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseUhOh.wav&quot;}</code></li>
<li><code>sound_rage</code> for <code>{&quot;FileName&quot;: &quot;s_Rage.wav&quot;}</code></li>
<li><code>sound_sadness_1</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness.wav&quot;}</code></li>
<li><code>sound_sadness_2</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness2.wav&quot;}</code></li>
<li><code>sound_sadness_3</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness3.wav&quot;}</code></li>
<li><code>sound_sadness_4</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness4.wav&quot;}</code></li>
<li><code>sound_sadness_5</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness5.wav&quot;}</code></li>
<li><code>sound_sadness_6</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness6.wav&quot;}</code></li>
<li><code>sound_sadness_7</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness7.wav&quot;}</code></li>
<li><code>sound_sleepy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy.wav&quot;}</code></li>
<li><code>sound_sleepy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy2.wav&quot;}</code></li>
<li><code>sound_sleepy_3</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy3.wav&quot;}</code></li>
<li><code>sound_sleepy_4</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy4.wav&quot;}</code></li>
<li><code>sound_sleepy_snore</code> for <code>{&quot;FileName&quot;: &quot;s_SleepySnore.wav&quot;}</code></li>
<li><code>sound_camera_shutter</code> for <code>{&quot;FileName&quot;: &quot;s_SystemCameraShutter.wav&quot;}</code></li>
<li><code>sound_failure</code> for <code>{&quot;FileName&quot;: &quot;s_SystemFailure.wav&quot;}</code></li>
<li><code>sound_success</code> for <code>{&quot;FileName&quot;: &quot;s_SystemSuccess.wav&quot;}</code></li>
<li><code>sound_wake</code> for <code>{&quot;FileName&quot;: &quot;s_SystemWakeWord.wav&quot;}</code></li>
</ul>
</details>
"""



"""The module contains the class Misty which represents a Misty II robot.
"""
from misty2py.action import *
from misty2py.information import *
from misty2py.misty_event import MistyEventHandler
from misty2py.response import *


class Misty:
    """A class representing a Misty robot.

    Attributes:
        ip (str): The IP address of this Misty robot.
        infos (Info): The object of Info class that belongs to this Misty.
        actions (Action): The object of Action class that belongs to this Misty.
        events (dict): A dictionary of active event subscriptions (keys being the event name, values the MistyEvent() object).
    """

    def __init__(
        self,
        ip: str,
        custom_info: Dict = {},
        custom_actions: Dict = {},
        custom_data: Dict = {},
        rest_protocol: str = "http",
        websocket_protocol: str = "ws",
        websocket_endpoint: str = "pubsub",
    ) -> None:
        """Initialises an instance of a Misty class representing a Misty robot.

        Args:
            ip (str): The IP address of a Misty robot.
            custom_info (Dict, optional): Custom information keywords in the form of dictionary with key being the keyword and value being the API endpoint that allows a `GET` request method. Defaults to {}.
            custom_actions (Dict, optional): Custom actions keywords in the form of dictionary with key being the keyword and value being a dictionary of the endpoint (key `"endpoint"`) and the method (key `"method"`) whose value is not `GET`. Defaults to {}.
            custom_data (Dict, optional): Custom data shortcuts in the form of dictionary with keys being the shortcuts and values being the json data in the form of a dictionary. Defaults to {}.
            rest_protocol (str, optional): The protocol to use when communicating with Misty's REST API. Defaults to "http".
            websocket_protocol (str, optional): The protocol to use when communicating with Misty's WebSocket API. Defaults to "ws".
            websocket_endpoint (str, optional): Misty's WebSocket API endpoint. Defaults to "pubsub".
        """
        self.ip = ip
        self.infos = Info(ip, rest_protocol, custom_allowed_infos=custom_info)
        self.actions = Action(
            ip,
            rest_protocol,
            custom_allowed_actions=custom_actions,
            custom_allowed_data=custom_data,
        )
        self.event_handler = MistyEventHandler(
            ip, websocket_protocol, websocket_endpoint
        )

    def __str__(self) -> str:
        """Parses the Misty object into a string.

        Returns:
            str: A string identifiyng the Misty object.
        """

        return "A Misty II robot with IP address %s" % self.ip

    def perform_action(self, action_name: str, data: Dict = {}) -> Misty2pyResponse:
        """Sends a request to perform an action to Misty's REST API.

        Args:
            action_name (str): The keyword specifying the action to perform.
            data (Dict, optional): The data to send in the request body in the form of a data shortcut or a json dictionary. Defaults to `{}`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        return self.actions.action_handler(action_name, data)

    def get_info(self, info_name: str, params: Dict = {}) -> Misty2pyResponse:
        """Sends a request to obtain information from Misty's REST API.

        Args:
            info_name (str): The information keyword specifying which information to retrieve.
            params (Dict): A dictionary of parameter names and parameter values. Defaults to `{}`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        return self.infos.get_info(info_name, params)

    def event(self, action: str, **kwargs) -> Misty2pyResponse:
        """Handles event-related actions by sending messeges to Misty's WebSocket API and receiving messages from it.

        Supports following actions:

        - **event subscripton** - requires an action keyword `"subscribe"` and an argument `type` (str) which is the name of the event type to subscribe to. Optional arguments are:
            - `name` (str) for a custom, unique event name.
            - `return_property` (str) for the property to return from Misty's WebSocket API. All properties are returned if `return_property` is not supplied.
            - debounce (int) for the interval at which new information is sent in ms. Defaults to `250`.
            - len_data_entries (int) for the maximum number of data entries to keep (discards in FIFO style). Defaults to `10`.
            - event_emitter (Callable) for an event emitter function which emits an event upon message recieval. Defaults to `None`.
        - **obtaining the data from an event** - requires an action keyword `"get_data"` and an argument `name` (str) specifying the event name.
        - **obtaining the log from an event** - requires an action keyword `"get_log"` and an argument `name` (str) specifying the event name.
        - **unsubscribing from an event** - requires an action keyword `"unsubscribe"` and an argument `name` (str) specifying the event name.

        Args:
            action (str): The action keyword. One of `"subscribe"`, `"get_data"`, `"get_log"` and `"unsubscribe"`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """

        if action == "subscribe":
            return self.event_handler.subscribe_event(kwargs)

        if action == "get_data":
            return self.event_handler.get_event_data(kwargs)

        if action == "get_log":
            return self.event_handler.get_event_log(kwargs)

        if action == "unsubscribe":
            return self.event_handler.unsubscribe_event(kwargs)

        return Misty2pyResponse(
            False,
            error_type=Misty2pyErrorType.COMMAND,
            error_msg="The event action `%s` is not supported." % action,
        )



"""This module enables speech with an extended return message.
"""
from typing import Dict

from misty2py.robot import Misty
from misty2py.utils.generators import get_random_string
from misty2py.response import compose_custom_response


def speak(misty: Misty, utterance: str) -> Dict:
    """Speaks an utterance and returns descriptive message.

    Args:
        misty (Misty): The Misty that speaks the utterance.
        utterance (str): The utterance to speak.

    Returns:
        Dict: the enhanced dictionarised Misty2pyResponse.
    """
    result = misty.perform_action(
        "speak",
        data={"Text": utterance, "UtteranceId": "utterance_" + get_random_string(6)},
    ).parse_to_dict()
    return compose_custom_response(
        result,
        success_message="Talking successful. Utterance: `%s`." % utterance,
        fail_message="Talking failed. Utterance: `%s`." % utterance,
    )



"""This module defines Misty2pyResponse class to handle responses from Misty2py when it communicates with Misty's REST API or Misty's WebSocket API.
"""

from typing import Any, Dict, List, Optional
from enum import Enum


class Misty2pyErrorType(Enum):
    """This class represents the types of errors originating in the Misty2py.
    """
    DATA_FORMAT = "incorrect data format"
    """Represents an error originating from the user supplying incorrect data format to any method, class or function."""
    DATA_SHORTCUT = "unsupported data shortcut"
    """Represents an error originating from the user supplying an unsupported data shortcut to any method, class or function."""
    UNKNOWN = "unknown"
    """Represents an error originating from the user supplying an unsupported command string to methods or functions that take commands as strings."""
    COMMAND = "unsupported command"
    """Represents an error originating from the user supplying unsupported HTTP request method or an unsupported command type to methods, classes or functions handling HTTP requests to Misty's REST API."""
    REQUEST_METHOD = "unsupported request method"
    """Represents an error whose origin is not known to the application."""
    MISSING = "missing argument or parameter"
    """Represents an error originating from missing arguments or parameters."""
    NONE = None
    """Represents that no error is present."""

    def __str__(self) -> str:
        """Parses an error type into a string."""
        return str(self.value)


class Misty2pyResponse:
    """A class representing a response from any function or method of Misty2py that communicates with Misty's REST API or Misty's WebSocket API.

    Attributes:
        misty2py_status (bool): Indicates whether any error originating in Misty2py was encountered.
        rest_response (Any): The response from Misty's REST API. `None` if this response does not contain a sub-response from Misty's REST API.
        ws_response (Optional[Dict]): The response from Misty's WebSocket API. `None` if this response does not contain a sub-response from Misty's WebSocket API.
        error_type (Misty2pyErrorType): The type of Misty2py error encountered. `Misty2pyErrorType.NONE` if no error in Misty2py was encountered.
        error_msg (Optional[str]): The error message for a Misty2py error. `None` if no error in Misty2py was encountered.
    """

    def __init__(
        self,
        misty2py_status: bool,
        rest_response: Any = None,
        ws_response: Optional[Dict] = None,
        error_type: Misty2pyErrorType = Misty2pyErrorType.NONE,
        error_msg: Optional[str] = None,
    ) -> None:
        """Creates a Misty2pyResponse object for a specific action.

        Args:
            misty2py_status (bool): Indicates whether any error originating in Misty2py was encountered.
            rest_response (Any, optional): The response from Misty's REST API. `None` if this response does not contain a sub-response from Misty's REST API. Defaults to `None`.
            ws_response (Optional[Dict], optional): The response from Misty's WebSocket API. `None` if this response does not contain a sub-response from Misty's WebSocket API. Defaults to `None`.
            error_type (Misty2pyErrorType, optional): The type of Misty2py error encountered. `Misty2pyErrorType.NONE` if no error in Misty2py was encountered. Defaults to `Misty2pyErrorType.NONE`.
            error_msg (Optional[str], optional): The error message for a Misty2py error. `None` if no error in Misty2py was encountered. Defaults to `None`.
        """
        self.misty2py_status = misty2py_status
        self.rest_response = rest_response
        self.error_type = error_type
        self.error_msg = error_msg
        self.ws_response = ws_response

    def parse_to_dict(self) -> Dict:
        """Parses this Misty2py object into a dictionary.

        Returns:
            Dict: A dictionary with keys `"misty2py_response"` (Dict) for the response of the Misty2py package, `"overall_success"` (bool) and optional keys `"rest_response"` (Dict) and `"ws_response"` (Dict).
        """
        dct = {
            "misty2py_response": self.get_misty2py_response(),
            "overall_success": self._is_successful(),
        }

        if self.rest_response is not None:
            dct["rest_response"] = self.get_rest_response()

        if self.ws_response is not None:
            dct["ws_response"] = self.ws_response

        return dct

    def get_rest_response(self) -> Dict:
        """Returns the REST API sub-response of this Misty2pyResponse with keys `"success"` and optional keys `"message"` or `"content"`."""
        if isinstance(self.rest_response, Dict):
            dct = self.rest_response.copy()
            st = dct.pop("status", None)
            if st == "Success":
                dct["success"] = True
            else:
                dct["success"] = False
            return dct

        return self.rest_response

    def get_ws_response(self) -> Dict:
        """Returns the WebSocket API sub-response of this Misty2pyResponse."""
        return self.ws_response

    def get_misty2py_response(self) -> Dict:
        """Parses the Misty2py part of this Misty2pyResponse to a dictionary containing keys `"success"` and, if the value of `"success"` is `False`, the `"error_msg"` (str) and `"error_type"` (str)."""
        success = self.misty2py_status
        dct = {"success": success}

        if not success:
            dct["error_msg"] = self.error_msg
            dct["error_type"] = str(self.error_type)

        return dct

    def _is_successful(self) -> bool:
        """Returns `True` if every sub-response present in this Misty2pyResponse is successful. Returns `False` otherwise."""
        success = self.misty2py_status
        if self.ws_response is not None:
            success = success and self.ws_response.get("success")
        if self.rest_response is not None:
            success = success and self.rest_response.get("status") == "Success"
        return success


def unknown_error(e: Exception) -> Misty2pyResponse:
    """Generates a Misty2pyResponse for an error `e` originating in unknown circumstances."""
    return Misty2pyResponse(False, error_msg=e, error_type=Misty2pyErrorType.UNKNOWN)


def success_of_action_list(msg_list: List[Dict[str, Dict]]) -> Dict:
    """Parses the overall success of a list of dictionaries where the keyw is an action name and the value is a dictionarised Misty2pyResponse. `overall_success` is only true if all actions were successful.

    Args:
        msg_list (List[Dict[str, Dict]]): The list of dictionaries where the keyword is an action name and the value is a Misty2pyResponse.

    Returns:
        Dict: The dictionary of the keys `"overall_success"` (bool) and `"actions"` whose value is a list of dictionaries with keys being action names and values being Misty2pyResponses.

    """
    dct = {"actions": []}
    success = True
    for event in msg_list:
        for name, message in event.items():
            if not message.get("overall_success"):
                success = False
        dct["actions"].append((name, message))
    dct["overall_success"] = success
    return dct


def success_of_action_dict(**messages) -> Dict:
    """Parses the overall success of a dictionary of actions, where the key is an action name and the value is a dictionarised Misty2pyResponse. `overall_success` is only true if all actions were successful.

    Returns:
        Dict: The dictionary of the keys `"overall_success"` (bool) and action names that contain the dictionarised Misty2pyResponses.
    """
    status_dict = {}
    overall_success = True
    for name, message in messages.items():
        status_dict[name] = message
        if not message.get("overall_success"):
            overall_success = False
    status_dict["overall_success"] = overall_success
    return status_dict


def compose_custom_response(
    resp: Dict,
    success_message: str = "Operation successful.",
    fail_message: str = "Operation failed.",
) -> Dict:
    """Enhances a dictionarised Misty2pyResponse with `success_message` in case of success and with `fail_message` otherwise.

    Args:
        resp (Dict): The dictionarised Misty2pyResponse.
        success_message (str, optional): A message/keyword to append in case of success. Defaults to `"Operation successful."`.
        fail_message (str, optional): A message/keyword to append in case of failure. Defaults to `"Operation failed."`.

    Returns:
        Dict: The enhanced version.
    """

    def compose_str(
        main_str: str, potential_str: Optional[str], fallback: Optional[str] = None
    ) -> str:
        """Composes a single string from main_str, potential_str and fallback.
        - If `potential_str` and `fallback` are both `None`, the final string is `main_str`.
        - If `potential_str` is a string, the final string is `main_str` followed by a space and `potential_str`.
        - If `potential_str` is `None` and `fallback` is a string, the final string is `main_str` followed by a space and `fallback`.
        """
        if isinstance(potential_str, str):
            return "%s %s" % (main_str, potential_str)
        if isinstance(fallback, str):
            return "%s %s" % (main_str, fallback)
        return main_str

    success = resp.get("overall_success")
    dct = {"overall_success": success}
    potential_resp = resp.get("rest_response")

    if potential_resp:
        potential_message = potential_resp.get("message")

        if success:
            message = compose_str(success_message, potential_message)
        else:
            message = compose_str(
                fail_message, potential_message, fallback="No further details provided."
            )
        dct["rest_response"] = {
            "success": potential_resp.get("success"),
            "message": message,
        }

    return dct



"""This module's function is to send information requests via the information keywords matching to Misty's API endpoints, sending information requests and matching data shortcuts.
"""
import json
from os import path
from typing import Dict

import requests

from misty2py.response import *

this_directory = path.abspath(path.dirname(__file__))
INFOS_JSON = str(path.join(this_directory, "allowed_infos.json"))


class Get:
    """A class representing the GET url request method.

    Attributes:
        ip (str): The IP address for the URL where the requests are sent.
        protocol (str): The protocol for the URL where the requests are sent.
        allowed_infos (dict): The dictionary of information keywords matching to the Misty's API endpoints.
    """

    def __init__(self, ip: str, protocol: str, custom_allowed_infos: Dict = {}) -> None:
        """Initialises a Get object.

        Args:
            ip (str): The IP address for the URL where the requests are sent.
            protocol (str): The protocol for the URL where the requests are sent.
            custom_allowed_infos (Dict, optional): The dictionary of custom information keywords. Defaults to `{}`.
        """

        self.ip = ip
        self.protocol = protocol

        allowed_infos = custom_allowed_infos
        f = open(INFOS_JSON)
        allowed_infos.update(json.loads(f.read()))
        f.close()

        self.allowed_infos = allowed_infos

    def get_info(self, endpoint: str, params: Dict) -> Misty2pyResponse:
        """Sends a GET request.

        Args:
            endpoint (str): The API endpoint to which the request is sent.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        r = requests.get(
            "%s://%s/%s" % (self.protocol, self.ip, endpoint), params=params
        )
        try:
            return Misty2pyResponse(True, rest_response=r.json())

        except Exception as e:
            return Misty2pyResponse(
                False,
                rest_response=r.content,
                error_msg=e,
                error_type=Misty2pyErrorType.UNKNOWN,
            )


class Info(Get):
    """A class representing an information request from Misty.
    A subclass of Get()."""

    def get_info(self, info_name: str, params: Dict = {}) -> Misty2pyResponse:
        """Sends an information request to Misty.

        Args:
            info_name (str): The information keyword specifying which information is requested.
            params (Dict): dict of parameter name and parameter value. Defaults to `{}`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        if info_name not in self.allowed_infos.keys():
            return Misty2pyResponse(
                False,
                error_msg="Command `%s` not supported." % info_name,
                error_type=Misty2pyErrorType.COMMAND,
            )

        endpoint = self.allowed_infos[info_name]

        try:
            return super().get_info(endpoint, params)

        except Exception as e:
            return unknown_error(e)



"""This module handles the communication of Misty2py with Misty's WebSocket API and the communication between different threads created within Misty2py for the WebSocket-related actions.
"""
import json
import threading
from typing import Callable, Dict, Union

import websocket

from misty2py.utils.generators import get_random_string
from misty2py.response import *


DEFAULT_DEBOUNCE = 250
DEFAULT_LEN_ENTRIES = 10


class MistyEvent:
    """A class that represents an event type subscribed to.
    Attributes:
        url (str): The URL of Misty's WebSocket API.
        data (list): The data entries obtained.
        type_str (str): The event type string as required by Misty's WebSockets API.
        event_name (str): A custom, unique event name.
        return_property (str): The property to return as requeired by Misty's WebSockets API.
        debounce (int): The interval at which new information is sent in ms.
        log (list): The logs.
        len_data_entries (int): The maximum number of data entries to keep.
        ee (Union[bool, Callable]): The event emitter function if one is desired, False otherwise.
    """

    def __init__(
        self,
        url: str,
        type_str: str,
        event_name: str,
        return_property: str,
        debounce: int,
        len_data_entries: int,
        event_emitter: Union[Callable, None],
    ):
        """Initialises an event object.
        Args:
            url (str): The URL of Misty's WebSocket API.
            type_str (str): The event type string as required by Misty's WebSockets API.
            event_name (str): A custom, unique event name.
            return_property (str): The property to return as required by Misty's WebSockets API.
            debounce (int): The interval at which new information is sent in ms.
            len_data_entries (int): The maximum number of data entries to keep.
            event_emitter (Union[Callable, None]): The event emitter function if one is desired, False otherwise.
        """
        self.url = url
        self.data = []
        self.type_str = type_str
        self.event_name = event_name
        self.return_property = return_property
        self.debounce = debounce
        self.log = []
        self.len_data_entries = len_data_entries
        event_thread = threading.Thread(target=self.run, daemon=True)
        event_thread.start()
        if event_emitter:
            self.ee = event_emitter
        else:
            self.ee = False

    def run(self):
        """Initialises the subscription and data collection."""
        self.ws = websocket.WebSocketApp(
            self.url,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.ws.run_forever()

    def on_message(self, ws, message):
        """Saves received data and if ee is set, emits an event."""

        message = json.loads(message)
        mes = message["message"]
        if len(self.data) > self.len_data_entries:
            self.data = self.data[1:-1]
        self.data.append(mes)

        if self.ee:
            self.ee.emit(self.event_name, mes)

    def on_error(self, ws, error):
        """Logs an error and if ee is set, emits an 'error' event."""

        if len(self.log) > self.len_data_entries:
            self.log = self.log[1:-1]
        self.log.append(error)

        if self.ee:
            self.ee.emit("error_%s" % self.event_name, error)

    def on_close(self, ws):
        """Appends the closing message to the log and if ee is set, emits a 'close' event."""

        mes = "Closed"
        if len(self.log) > self.len_data_entries:
            self.log = self.log[1:-1]
        self.log.append(mes)

        if self.ee:
            self.ee.emit("close_%s" % self.event_name, mes)

    def on_open(self, ws):
        """Appends the opening message to the log and starts the subscription and if ee is set, emits an 'open' event."""

        self.log.append("Opened")
        self.subscribe()
        ws.send("")

        if self.ee:
            self.ee.emit("open_%s" % self.event_name)

    def subscribe(self):
        """Constructs the subscription message."""

        msg = {
            "Operation": "subscribe",
            "Type": self.type_str,
            "DebounceMs": self.debounce,
            "EventName": self.event_name,
            "ReturnProperty": self.return_property,
        }
        msg_str = json.dumps(msg, separators=(",", ":"))
        self.ws.send(msg_str)

    def unsubscribe(self):
        """Constructs the unsubscription message."""

        msg = {"Operation": "unsubscribe", "EventName": self.event_name, "Message": ""}
        msg_str = json.dumps(msg, separators=(",", ":"))
        self.ws.send(msg_str)
        self.ws.close()


class MistyEventHandler:
    """A class that handles all events its related Misty object subscribed to during this runtime.

    Attributes:
        events (Dict): The dictionary of all Event objects their related Misty object subscribed to during the current runtime.
        url (str): The URL for Misty's WebSocket API.
    """

    def __init__(self, ip: str, protocol: str, endpoint: str) -> None:
        """Initialises an object of class MistyEventHandler.

        Args:
            ip (str): The IP address for the URL where the requests are sent.
            protocol (str): The protocol for the URL where the requests are sent.
            endpoint (str): The endpoint for the URL where the requests are sent.
        """

        self.events = {}
        self.url = "%s://%s/%s" % (protocol, ip, endpoint)

    def subscribe_event(self, kwargs: Dict) -> Misty2pyResponse:
        """Subscribes to an event type.

        Args:
            kwargs (Dict):  requires a key `"type"` (a string representing the event type to subscribe to). Optional keys are:
            - `name` (str) for a custom event name; must be unique.
            - `return_property` (str) for the property to return from Misty's websockets; all properties are returned if return_property is not supplied.
            - `debounce` (int) for the interval at which new information is sent in ms; defaults to `250`.
            - `len_data_entries` (int) for the maximum number of data entries to keep (discards in fifo style); defaults to `10`.
            - `event_emitter` (Callable) for an event emitter function which emits an event upon message recieval.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """

        event_type = kwargs.get("type")

        if not event_type:
            return Misty2pyResponse(
                False,
                error_msg="No event type specified.",
                error_type=Misty2pyErrorType.MISSING,
            )
        if not isinstance(event_type, str):
            return Misty2pyResponse(
                False,
                error_msg="Not a valid event type: `%s`." % str(event_type),
                error_type=Misty2pyErrorType.DATA_FORMAT,
            )

        event_name = kwargs.get("name")
        if not event_name:
            event_name = "event_%s_%s" % (event_type, get_random_string(8))

        return_property = kwargs.get("return_property")

        debounce = kwargs.get("debounce")
        if not debounce:
            debounce = DEFAULT_DEBOUNCE

        len_data_entries = kwargs.get("len_data_entries")
        if not len_data_entries:
            len_data_entries = DEFAULT_LEN_ENTRIES

        event_emitter = kwargs.get("event_emitter")

        try:
            new_event = MistyEvent(
                self.url,
                event_type,
                event_name,
                return_property,
                debounce,
                len_data_entries,
                event_emitter,
            )

        except Exception as e:
            return unknown_error(e)

        self.events[event_name] = new_event

        return Misty2pyResponse(
            True,
            ws_response={
                "success": True,
                "event_name": event_name,
                "message": "Subscribed to event type `%s` with name `%s`"
                % (event_type, event_name),
            },
        )

    def _obtain_event_data_or_log(
        self, event_name: Optional[str], is_log: bool = False
    ) -> Misty2pyResponse:
        """Obtains event data or event log.

        Args:
            event_name (Optional[str]): The name of the event whose data or log to obtain.
            is_log (bool, optional): `False` to obtain data, `True` to obtain log. Defaults to `False`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """
        if not event_name:
            return Misty2pyResponse(
                False,
                error_msg="No event name specified.",
                error_type=Misty2pyErrorType.MISSING,
            )

        if event_name in self.events.keys():
            try:
                if is_log:
                    return Misty2pyResponse(
                        True,
                        ws_response={
                            "success": True,
                            "message": self.events[event_name].log,
                        },
                    )

                return Misty2pyResponse(
                    True,
                    ws_response={
                        "success": True,
                        "message": self.events[event_name].data,
                    },
                )

            except Exception as e:
                return Misty2pyResponse(
                    True,
                    ws_response={
                        "success": False,
                        "message": "Error occurred while attempting to unsubscribe from event `%s`. Error message: `%s`."
                        % (event_name, e),
                    },
                )

        return Misty2pyResponse(
            True,
            ws_response={
                "success": False,
                "message": "Event type `%s` is not subscribed to." % event_name,
            },
        )

    def get_event_data(self, kwargs: Dict) -> Misty2pyResponse:
        """Obtains data from a subscribed event type.

        Args:
            kwargs (Dict): Requires a key "name" (the event name).

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """

        return self._obtain_event_data_or_log(kwargs.get("name"))

    def get_event_log(self, kwargs: Dict) -> Misty2pyResponse:
        """Obtains the log from a subscribed event type.

        Args:
            kwargs (Dict): Requires a key `name` (the event name).

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """

        return self._obtain_event_data_or_log(kwargs.get("name"), is_log=True)

    def unsubscribe_event(self, kwargs: Dict) -> Misty2pyResponse:
        """Unsubscribes from an event type.

        Args:
            kwargs (Dict): Requires a key `name` (the event name).

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty WebSocket API sub-response.
        """

        event_name = kwargs.get("name")
        if not event_name:
            return Misty2pyResponse(
                False,
                error_msg="No event name specified.",
                error_type=Misty2pyErrorType.MISSING,
            )

        if event_name in self.events.keys():
            try:
                self.events[event_name].unsubscribe()

            except Exception as e:
                return Misty2pyResponse(
                    True,
                    ws_response={
                        "success": False,
                        "message": "Error occurred while attempting to unsubscribe from event `%s`. Error message: `%s`."
                        % (event_name, e),
                    },
                )

            resp = Misty2pyResponse(
                True,
                ws_response={
                    "success": True,
                    "message": "Event `%s` of type `%s` unsubscribed"
                    % (event_name, self.events[event_name].type_str),
                    "log": self.events[event_name].log,
                },
            )
            self.events.pop(event_name)
            return resp

        return Misty2pyResponse(
            True,
            ws_response={
                "success": False,
                "message": "Event type `%s` is not subscribed to." % event_name,
            },
        )



"""This module's function is to send action requests via the action keywords matching to Misty's API endpoints, sending action requests and matching data shortcuts.
"""
import json
from os import path
from typing import Dict, Union

import requests

from misty2py.utils.colours import construct_transition_dict
from misty2py.response import *

this_directory = path.abspath(path.dirname(__file__))
ACTIONS_JSON = str(path.join(this_directory, "allowed_actions.json"))
DATA_JSON = str(path.join(this_directory, "allowed_data.json"))
VALID_HTTP_REQUEST_METHODS = ["OPTIONS", "HEAD", "POST", "PUT", "PATCH", "DELETE"]


class BodyRequest:
    """A class representing the url request methods with a body.

    Attributes:
        ip (str): The IP address for the URL where the requests are sent.
        protocol (str): The protocol for the URL where the requests are sent.
        allowed_actions (dict): The dictionary of custom action keywords matching to the Misty's REST API endpoints.
        allowed_data (dict): The dictionary of custom data shortcuts matching to the json dictionaries required by Misty's REST API.
    """

    def __init__(
        self,
        ip: str,
        protocol: str,
        custom_allowed_actions: Dict = {},
        custom_allowed_data: Dict = {},
    ) -> None:
        """Initialises a Post object.

        Args:
            ip (str): The IP address where the requests are sent.
            protocol (str): The protocol for the URL where the requests are sent.
            custom_allowed_actions (Dict, optional): The dictionary of action keywords. Defaults to `{}`.
            custom_allowed_data (Dict, optional): The dictionary of data shortcuts. Defaults to `{}`.
        """

        self.ip = ip
        self.protocol = protocol

        allowed_actions = custom_allowed_actions
        f = open(ACTIONS_JSON)
        allowed_actions.update(json.loads(f.read()))
        f.close()
        self.allowed_actions = allowed_actions

        allowed_data = custom_allowed_data
        f = open(DATA_JSON)
        allowed_data.update(json.loads(f.read()))
        f.close()
        self.allowed_data = allowed_data

    def perform_action(
        self, endpoint: str, data: Dict, request_method: str = "post"
    ) -> Misty2pyResponse:
        """Sends an action request.

        Args:
            endpoint (str): The REST API endpoint to which the request is sent.
            data (Dict): The json data supplied in the body of the request.
            request_method (str, optional): The request method. Defaults to `"post"`.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        if request_method.upper() not in VALID_HTTP_REQUEST_METHODS:
            return Misty2pyResponse(
                False,
                error_msg="Request method `%s` is not supported." % request_method,
                error_type=Misty2pyErrorType.REQUEST_METHOD,
            )

        response = requests.request(
            request_method.upper(),
            "%s://%s/%s" % (self.protocol, self.ip, endpoint),
            json=data,
        )

        try:
            return Misty2pyResponse(True, rest_response=response.json())

        except Exception as e:
            return Misty2pyResponse(
                False,
                rest_response=response.content,
                error_msg=e,
                error_type=Misty2pyErrorType.UNKNOWN,
            )


class Action(BodyRequest):
    """A class representing an action request for Misty."""

    def perform_action(
        self, action_name: str, data: Union[str, Dict], data_method: str
    ) -> Misty2pyResponse:
        """Sends an action request to Misty.

        Args:
            action_name (str): The action keyword specifying which action is requested.
            data (Union[str, Dict]): The data shortcut representing the data supplied in the body of the request or the json dictionary to be supplied in the body of the request.
            data_method (str): "dict" if the data is supplied as a json dictionary, "string" if the data is supplied as a data shortcut.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        if action_name not in self.allowed_actions.keys():
            return Misty2pyResponse(
                False,
                error_msg="Command `%s` not supported." % action_name,
                error_type=Misty2pyErrorType.COMMAND,
            )

        if data_method == "dict":
            try:
                return super().perform_action(
                    self.allowed_actions[action_name]["endpoint"],
                    data,
                    request_method=(self.allowed_actions[action_name]["method"]),
                )

            except Exception as e:
                return unknown_error(e)

        if data_method == "string" and data in self.allowed_data:
            try:
                return super().perform_action(
                    self.allowed_actions[action_name]["endpoint"],
                    self.allowed_data[data],
                    request_method=(self.allowed_actions[action_name]["method"]),
                )

            except Exception as e:
                return unknown_error(e)

        else:
            return Misty2pyResponse(
                False,
                error_type=Misty2pyErrorType.DATA_SHORTCUT,
                error_msg="Data shortcut `%s` is not supported." % data,
            )

    def action_handler(
        self, action_name: str, data: Union[Dict, str]
    ) -> Misty2pyResponse:
        """Sends Misty a request to perform an action.

        Args:
            action_name (str): The keyword specifying the action to perform.
            data (Union[Dict, str]): The data to send in the request body in the form of a data shortcut or a json dictionary.

        Returns:
            Misty2pyResponse: A Misty2pyResponse object with Misty2py sub-response and Misty REST API sub-response.
        """

        if (
            action_name == "led_trans"
            and isinstance(data, Dict)
            and len(data) >= 2
            and len(data) <= 4
        ):

            try:
                data = construct_transition_dict(data, self.allowed_data)

            except ValueError as e:
                return Misty2pyResponse(
                    False, error_msg=e, error_type=Misty2pyErrorType.DATA_FORMAT
                )

        data_method = "string"

        if isinstance(data, Dict):
            data_method = "dict"

        return self.perform_action(action_name, data, data_method)
